# decotrace - Function call/return value logging.
# Copyright (C) 2011 Nathan Wilcox
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
decotrace provides a function, method, and class decorator which sends
messages about call arguments, return values, and exceptions to an
arbitrary output function.

The default output function is to write to sys.stderr, but a common use
case is to provide a logging context.

Example usage:

>>> from decotrace import traced
>>> @traced
... def f(x):
...   return (x, x)
... 
>>> f(42)
f calling with (42)...
f return (42, 42)
(42, 42)

Using a custom output:
>>> import sys, logging, decotrace
>>> logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
>>> traced = decotrace.TraceContext(logging.getLogger('-decotrace-').debug)
>>> @traced
... def f(x):
...   return x * 2
... 
>>> f(7)
DEBUG:-decotrace-:f calling with (7)...
DEBUG:-decotrace-:f return 14
14
"""

import sys
import inspect
from types import ClassType, FunctionType, MethodType


class TraceContext (object):
    def __init__(self, output = lambda s: sys.stderr.write(s + '\n')):
        self._output = output
        self._depth = 0

    def __call__(self, thing):
        t = type(thing)
        if t is FunctionType:
            return self._wrap_funclike(thing)
        elif t is MethodType:
            return self._wrap_function(thing)
        elif t in [type, ClassType]:
            return self._wrap_type(thing)
        else:
            raise NotImplementedError('Cannot wrap: {0!r}'.format(t))
        
    # Private:
    def _log(self, tmpl, *args, **kw):
        prefix = '  ' * self._depth
        self._output('{0}{1}'.format(prefix, tmpl.format(*args, **kw)))

    def _wrap_funclike(self, c):
        (argnames, _, _, _) = inspect.getargspec(c)
        if argnames and argnames[0] == 'self':
            return self._wrap_method(c)
        else:
            return self._wrap_function(c)

    def _wrap_method(self, m):
        def wrapped(wrappedSelf, *args, **kw):
            name = '{0}[@{1:08x}].{2}'.format(
                wrappedSelf.__class__.__name__,
                id(self),
                m.__name__)
            fullArgs = (wrappedSelf,) + args
            return self._trace_call(name, 1, m, fullArgs, kw)
        return wrapped

    def _wrap_function(self, f):
        name = f.__name__
        return lambda *a, **kw: self._trace_call(name, 0, f, a, kw)

    def _wrap_type(self, basetype):
        classdict = {}
        for n, v in vars(basetype).items():
            if type(v) is FunctionType:
                v = self(v)
            classdict[n] = v
        if type(basetype) is ClassType:
            bases = (basetype, object)
        else:
            bases = (basetype,)
        return type(basetype.__name__, bases, classdict)

    def _trace_call(self, name, argStart, f, args, kw):
        argDesc = self._describeArgs(args[argStart:], kw)
        self._log('{0}{1}...', name, argDesc)
        try:
            self._depth += 1
            try:
                result = f(*args, **kw)
            finally:
                self._depth -= 1
        except Exception, e:
            self._log('{0}{1} raising {2!r}', name, argDesc, e)
            raise
        else:
            self._log('{0}{1} -> {2!r}', name, argDesc, result)
            return result

    @staticmethod
    def _describeArgs(args, kw):
        fields = [ repr(a) for a in args ]
        fields.extend( [ '{0}={1!r}'.format(k, v) for (k, v) in sorted(kw.items()) ] )
        return '({0})'.format(', '.join(fields))


traced = TraceContext()



#! /usr/bin/env python

from decotrace import traced

@traced
def add(a, b):
    return a + b

@traced
def recursive_geometric(n):
    if n == 0:
        return 0
    else:
        return n + recursive_geometric(n-1)

class Adder (object):
    def __init__(self, a):
        self.a = a

    @traced
    def add(self, b):
        return self.a + b

@traced
class Incrementer (int):
    def __init__(self, v=0):
        self.v = 0

    def inc(self, amt=1):
        self.v += amt
        return self.v

    classvar = 42

    def _p1(self):
        return self.classvar

    def __p2(self):
        return self.inc

    def __add__(self, other):
        return self.v + other



add(2, 3)
Adder(2).add(3)
recursive_geometric(3)
add(2, b=3)
add(a=2, b=3)
recursive_geometric(n=3)

i = Incrementer()
i.inc()
i.inc(amt=7)



#! /usr/bin/env python

import unittest
import re

import decotrace


class DecoTraceTests (unittest.TestCase):

    def setUp(self):
        self._trace_log = []
        self.traced = decotrace.TraceContext(self._trace_log.append)


    def _define_next_int(self):

        @self.traced
        def next_int(x):
            return x+1

        return next_int


    def test_simple_function(self):

        next_int = self._define_next_int()

        self.assertEqual(3, next_int(2))
        self.checkTraceLog(
            r'next_int\(2\)...',
            r'next_int\(2\) -> 3',
            )


    def test_keyword_call_display(self):

        next_int = self._define_next_int()

        self.assertEqual(3, next_int(x=2))
        self.checkTraceLog(
            r'next_int\(x=2\)...',
            r'next_int\(x=2\) -> 3',
            )


    def test_exception(self):

        next_int = self._define_next_int()

        try:
            r = next_int('x')
        except TypeError:
            pass
        else:
            self.fail("Expected TypeError, but instead next_int('x') returns {0!r}".format(r))

        self.checkTraceLog(
            r"next_int\('x'\)...",
            r"next_int\('x'\) raising TypeError\(\"cannot concatenate 'str' and 'int' objects\",\)",
            )


    def test_name_collisions(self):

        @self.traced
        def hello(name):
            return 'Hello, ' + name
        
        self.assertEqual('Hello, bob', hello(name='bob'))
        self.checkTraceLog(
            r"hello\(name='bob'\)...",
            r"hello\(name='bob'\) -> 'Hello, bob'",
            )


    def test_method_decoration(self):

        class C (object):
            @self.traced
            def method(self, x):
                return x*2

        i = C()
        result = i.method(3)

        self.assertEqual(6, result)
        self.checkTraceLog(
            r"C\[@[0-9a-f]{8}\].method\(3\)...",
            r"C\[@[0-9a-f]{8}\].method\(3\) -> 6",
            )


    def checkTraceLog(self, *entries):
        self.assertEqual(len(entries), len(self._trace_log))

        for expected, actual in zip(entries, self._trace_log):
            pat = re.compile(expected)
            m = pat.match(actual)
            self.failIf(m is None, '{0!r} does not match {1!r}'.format(expected, actual))


if __name__ == '__main__':
    unittest.main()



#! /usr/bin/env python

from setuptools import setup

setup(name = 'decotrace',
      description = 'Function/method/class call tracing decorator.',
      author = 'Nathan Wilcox',
      author_email = 'nejucomo@gmail.com',
      version = '0.1.2',
      url = 'https://bitbucket.org/nejucomo/decotrace',
      license = 'GPLv3',

      py_modules = ['decotrace'],

      test_suite = 'test_decotrace',

      setup_requires = ['setuptools_hg'],
      )
